/*
**	GetFT2.c
**
**	Copyright (C) 1995 Bernardo Innocenti
**
**	Load a FastTracker ][ module with any number of tracks.
*/

#include <exec/memory.h>
#include <libraries/XModule.h>

#include <clib/exec_protos.h>
#include <clib/dos_protos.h>
#include <clib/xmodule_protos.h>

#include <pragmas/exec_sysbase_pragmas.h>
#include <pragmas/dos_pragmas.h>
#include <pragmas/xmodule_pragmas.h>

#include "Gui.h"
#include "FastTracker2.h"


/* Convert an Intel style WORD to Motorola format */
#define I2M(x) ( (UWORD) ( (((UWORD)(x)) >> 8) | (((UWORD)(x)) << 8) ) )

/* Convert an Intel style LONG to Motorola format */
#define I2ML(x) ( I2M((x)>>16) | (I2M((x))<<16) )



LONG GetFT2 (struct SongInfo *si, BPTR fh)
{
	ULONG i, j;		/* Loop counters */
	struct FT2Header ft2hd;


	/* Read module header */
	if (Read (fh, &ft2hd, sizeof (ft2hd)) != sizeof (ft2hd))
		return ERR_READWRITE;

	/* Convert from Intel's shitty endian */
	ft2hd.HeaderSize	= I2ML(ft2hd.HeaderSize);
	ft2hd.Length		= I2M(ft2hd.Length);
	ft2hd.Restart		= I2M(ft2hd.Restart);
	ft2hd.Channels		= I2M(ft2hd.Channels);
	ft2hd.NumPatt		= I2M(ft2hd.NumPatt);
	ft2hd.NumInstr		= I2M(ft2hd.NumInstr);
	ft2hd.Flags			= I2M(ft2hd.Flags);
	ft2hd.DefTempo		= I2M(ft2hd.DefTempo);
	ft2hd.DefBPM		= I2M(ft2hd.DefBPM);

	if (!xmSetSongLen (si, ft2hd.Length))
		return ERROR_NO_FREE_STORE;

	if (Read (fh, si->Sequence, si->Length * sizeof (UWORD)) !=
		si->Length * sizeof (UWORD))
		return ERR_READWRITE;

	for (i = 0; i < si->Length; i++)
		si->Sequence[i] = I2M(si->Sequence[i]);

	si->Restart = ft2hd.Restart;
	si->GlobalSpeed = ft2hd.DefTempo;
	si->GlobalTempo = ft2hd.DefBPM;

	return RETURN_OK;
}
