/*
**	OktalyzerHook.c
**
**	Copyright (C) 1993,94,95,96,97 Bernardo Innocenti
**
**	Originally based on Gerardo Iula's Tracker sources.
**
**	External hook for Oktalyzer 1.1-1.57 module format.
*/


#include <exec/memory.h>
#include <dos/stdio.h>
#include <libraries/xmodule.h>
#include <libraries/songclass.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/xmodule.h>

#include "XModulePriv.h"
#include "Gui.h"



/* Local functions prototypes */

HOOKCALL APTR _GetEngine (
	REG(a6, struct Library *MyBase));
HOOKCALL APTR _SetupXMHook (
	REG(a0, struct XModuleBase *_XModuleBase),
	REG(a6, struct Library *mybase));
static HOOKCALL struct XMHook *IdentifyOktalyzer (
	REG(d0, BPTR fh),
	REG(a0, struct XMHook *loader),
	REG(a1, ULONG *tags));
static HOOKCALL LONG LoadOktalyzer (
	REG(d0, BPTR fh),
	REG(a0, struct SongInfo *si),
	REG(a1, struct XMHook *loader),
	REG(a2, ULONG *tags));
static HOOKCALL LONG SaveOktalyzer (
	REG(d0, BPTR fh),
	REG(a0, struct SongInfo *si),
	REG(a1, struct XMHook *saver),
	REG(a2, ULONG *tags));

INLINE UBYTE DecodeEff (UBYTE eff, UBYTE effval, UWORD patt, UWORD line, UWORD track);



#define OKT_MODE4	1	/* Mode 4: play 8 bit instruments */
#define OKT_MODE8	0	/* Mode 8: play 7 bit instruments */
#define OKT_MODEB	2	/* Mode B: play both 8 & 7 bit instruments */


/* Oktalyzer chunk IDs */

#define ID_OKTA 'OKTA'
#define ID_CMOD 'CMOD'
#define ID_SAMP	'SAMP'
#define ID_SPEE	'SPEE'
#define ID_SLEN	'SLEN'
#define ID_PLEN	'PLEN'
#define ID_PBOD 'PBOD'
#define ID_SBOD 'SBOD'

#ifndef ID_SONG
#define ID_SONG 'SONG'
#endif /* !ID_SONG */

#ifndef ID_PATT
#define ID_PATT 'PATT'
#endif /* !ID_PATT */



/* Library data */

const UBYTE LibName[] = "oktalyzer.xmhook";
const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
const UBYTE LibId[] = "oktalyzer.xmhook 1.0 (28.1.96)  1993-96 by Bernardo Innocenti";



/* Effects conversion table
 * Originally based on Gerardo Iula's "Tracker" source.
 */
static const UBYTE Effects[MAXTABLEEFFECTS] =
{
/*  OKTA		XModule			Val */

	0x00,	/*	Null effect		$00	*/

	0x01,	/*	Portamento Up	$01	*/
	0x02,	/*	Portamento Down	$02	*/
	0x00,	/*	Tone Portamento	$03	*/
	0x00,	/*	Vibrato			$04	*/
	0x00,	/*	ToneP + VolSl	$05	*/
	0x00,	/*	Vibra + VolSl	$06	*/
	0x00,	/*	Tremolo			$07	*/
	0x00,	/*	Set Hold/Decay	$08	*/
	0x00,	/*	Sample Offset	$09	*/
	0x1E,	/*	Volume Slide	$0A	*/
	0x19,	/*	Position Jump  	$0B	*/
	0x1F,	/*	Set Volume		$0C	*/
	0x00,	/*	Pattern break	$0D	*/
	0x00,	/*	Misc			$0E	*/
	0x1C,	/*	Set Speed		$0F	*/
	0x00,	/*	Set Tempo		$10	*/
	0x1A,	/*	Arpeggio		$11	*/

	0x11,	/*	Oktalyzer H        	*/
	0x15	/*	Oktalyzer L        	*/
};



/* Get around a SAS/C bug which causes some annoying warnings
 * with the library bases defined below.  The problem happens
 * when the GST has been compiled with the CODE=FAR switch and
 * the source is being compiled with CODE=NEAR.
 */
#ifdef __SASC
	#pragma msg 72 ignore push
#endif /* __SASC */

struct ExecBase			*SysBase		= NULL;
struct XModuleBase		*XModuleBase	= NULL;
struct DosLibrary		*DOSBase		= NULL;
struct IntuitionBase	*IntuitionBase	= NULL;

#ifdef __SASC
#pragma msg 72 pop
#endif /* __SASC */



static HOOKCALL struct XMHook *IdentifyOktalyzer (
	REG(d0, BPTR fh),
	REG(a0, struct XMHook *loader),
	REG(a1, ULONG *tags))

/* Determine if the given file is an Oktalyzer module.
 * Note: the file position will be changed on exit.
 */
{
	ULONG id[3];

	Seek (fh, 0, OFFSET_BEGINNING);
	if (FRead (fh, &id, 12, 1) != 1)
		return NULL;

	if ((id[0] == ID_OKTA) && (id[1] == ID_SONG) && (id[2] == ID_CMOD))
		return loader;

	return NULL;
}



INLINE UBYTE DecodeEff (UBYTE eff, UBYTE effval, UWORD patt, UWORD line, UWORD track)

/* Inputs: old effect & old type.
 * Output: new effect in requested newtype.
 */
{
	UBYTE i;

	for ( i = 0 ; i < MAXTABLEEFFECTS ; i++ )
		if (eff == Effects[i])
			return i;

	xmDisplayMessage (XMDMF_NOTE | XMDMF_USECATALOG,
		(APTR)MSG_UNKNOWN_EFF, 5, eff, patt, track, line);
	return 0;
}



static HOOKCALL LONG LoadOktalyzer (
	REG(d0, BPTR fh),
	REG(a0, struct SongInfo *si),
	REG(a1, struct XMHook *loader),
	REG(a2, ULONG *tags))
{
	struct Instrument	*instr;
	struct Pattern		*patt;
	ULONG i, j, k;	/* Loop counters	*/
	ULONG size;		/* Read buffer		*/
	ULONG l;		/* Read buffer		*/
	UWORD voices, numpatts, songlen, instr_mode[37];
	UWORD w;
	UBYTE oktanote[4];


	/* Check file header OKTASONGCMOD */
	{
		LONG id[3];

		if (FRead (fh, id, 12, 1) != 1) return ERROR_IOERR;
		if ((id[0] != ID_OKTA) || (id[1] != ID_SONG) || (id[2] != ID_CMOD))
			return ERROR_NOTMODULE;
	}


	/* TODO: set si->SongName */


	/* CMOD Chunk size ($0000 0008) */
	Seek (fh, 4, OFFSET_CURRENT); /* Skip 4 bytes */

	voices = 4;		/* Set minimum voices and check others	*/
	for (i = 0 ; i < 4 ; i++)
	{
		if (FRead (fh, &w, 2, 1) != 1) return ERROR_IOERR;
		if (w) voices++;
	}

	/* Get Sample Name, Length, Repeat Replen, Volume for each instr. */

	/* Check header SAMP */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
	if (l != ID_SAMP) return ERROR_NOTMODULE;

	/* SAMP Chunk Size ($0000 0480) */
	Seek (fh, 4, OFFSET_CURRENT); /* Skip 4 bytes */

	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_READING_INSTS_INFO, NULL);

	/* Read in 36 instruments */
	for ( j = 1 ; j <= 36 ; j++ )
	{
		struct
		{
			UBYTE	name[20];
			ULONG	size;
			UWORD	repeat,
					replen,
					volume,
					mode;
		} oktainstr;

		/* Get instrument data */
		if (FRead (fh, &oktainstr, sizeof (oktainstr), 1) != 1)
			return ERROR_IOERR;

		oktainstr.name[19] = '\0';

		/* Oktalyzer sometimes saves odd lengths for instruments,
		 * but the data saved in SBOD is always rounded _down_ to an even
		 * number!  It took me two weeks to find and kill this bug :-(
		 */
		oktainstr.size &= (~1);

		/* Oktalyzer instrument modes:
		 *	mode 8 $00 7 bit instruments
		 *	mode 4 $01 normal instruments
		 *	mode B $02 both 8bit & 7bit
		 */
		instr_mode[j] = oktainstr.mode;

		if (oktainstr.size)
			if (!(xmAddInstrument (si, j,
				INSTRA_Name,	oktainstr.name,
				INSTRA_Length,	oktainstr.size,
				INSTRA_Repeat,	oktainstr.repeat << 1,
				INSTRA_Replen,	oktainstr.replen << 1,
				INSTRA_Volume,	oktainstr.volume,
				TAG_DONE)))
				return ERROR_NO_FREE_STORE;
	}


	/* Get global song speed */

	/* Check speed header "SPEE" */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
	if (l != ID_SPEE) return ERROR_NOTMODULE;

	/* SPEE Chunk size ($0000 0002) */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	/* Get Song Global Speed */
	if (FRead (fh, &w, 2, 1) != 1) return ERROR_IOERR;

	SetAttrs (si,
		SNGA_GlobalSpeed, w,
		TAG_DONE);


	/* Get number of patterns */

	/* Check header SLEN */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
	if (l != ID_SLEN) return ERROR_NOTMODULE;

	/* SLEN Chunk size ($0000 0002) */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	/* Get number of patterns */
	if (FRead (fh, &numpatts, 2, 1) != 1) return ERROR_IOERR;

	/* Check value */
	if (numpatts > MAXPATTERNS)
	{
		xmDisplayMessageA (XMDMF_WARNING | XMDMF_USECATALOG,
			(APTR)MSG_SONG_HAS_TOO_MANY_PATT, NULL);
		numpatts = MAXPATTERNS - 1;
	}


	/* Get number of positions in song (Length) */

	/* Check header PLEN */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
	if (l != ID_PLEN) return ERROR_NOTMODULE;

	/* Chunk size ($0000 0002) */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	/* Get number of patterns */
	if (FRead (fh, &songlen, 2, 1) != 1) return ERROR_IOERR;

	/* Check value */
	if (songlen > MAXPOSITIONS)
	{
		xmDisplayMessageA (XMDMF_WARNING | XMDMF_USECATALOG,
			(APTR)MSG_SONG_TOO_LONG, NULL);
		songlen = MAXPOSITIONS;
	}


	/* Get position table */

	/* Check header PATT */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
	if (l != ID_PATT) return ERROR_NOTMODULE;

	/* PATT Chunk size ($0000 0080) */
	if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;


	/* Read song sequence */
	{
		UBYTE postable[128];

		if (FRead (fh, postable, 1, 128) != 128) return ERROR_IOERR;

		if (!(xmSetSongLen (si, songlen)))
			return ERROR_NO_FREE_STORE;

		for (i = 0; i < si->Length; i++)
			si->Sequence[i] = postable[i];
	}


	/* Get pattern bodies and convert them */

	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_READING_PATTS, NULL);

	for (j = 0; j < numpatts; j++)
	{
		if (xmDisplayProgress (j + 1, numpatts))
			return ERROR_BREAK;

		/* Check header "PBOD" */
		if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
		if (l != ID_PBOD) return ERROR_NOTMODULE;

		/* Skip Chunk Length (Lines * Tracks * 4 + 2) */
		if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;

		/* Get pattern length */
		if (FRead (fh, &w, 2, 1) != 1) return ERROR_IOERR;

		/* Allocate memory for all tracks */
		if (!(patt = xmAddPattern (si,
			PATTA_Tracks,	voices,
			PATTA_Lines,	w,
			TAG_DONE)))
			return ERROR_NO_FREE_STORE;

		for ( k = 0 ; k < patt->Lines ; k++)
		{
			for (i = 0; i < voices; i++)
			{
				struct Note *n = &patt->Notes[i][k];

				/* Read a whole track row */
				if (FRead (fh, oktanote, 4, 1) != 1) return ERROR_IOERR;

				/* Convert Note:
				 *
				 * Oktalyzer supports 3 octaves (1 to 3).
				 * Notes are numbered from 1 (C-1) to 36 (B-3).
				 * 0 means no note.
				 */
				if (oktanote[0] <= 36)
					n->Note = (oktanote[0] ? (oktanote[0] + 12) : 0);	/* Add one octave */
				else
					xmDisplayMessage (XMDMF_NOTE | XMDMF_USECATALOG,
						(APTR)MSG_INVALID_NOTE, oktanote[0], j, i, k);

				/* Store Instrument Number */
				n->Inst = (n->Note ? (oktanote[1] + 1) : 0);

				/* Convert Effect */
				n->EffNum = DecodeEff (oktanote[2], oktanote[3], j, k, i);

				/* Store Effect Value */
				n->EffVal = oktanote[3];

				/* Effect Exceptions */
				if (n->EffNum == EFF_MISC)
				{
					/* Oktalyzer has SetFilter values inverted! */
					if ((n->EffVal >> 4) == 1)
						n->EffVal = 0x10 | (n->EffVal & 0x0F ? 0 : 1);
				}
			}	/* end for (i) */
		}	/* End for (k) */
	}	/* End for (j) */


	/* Load instruments */

	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_READING_INSTS, NULL);

	for (j = 1 ; j <= 36 ; j++)
	{
		BYTE *sample;

		if (!(instr = si->Instr[j])) continue;

		if (xmDisplayProgress(j, 36))
			return ERROR_BREAK;

		/* Check header SBOD */
		if (FRead (fh, &l, 4, 1) != 1) return ERROR_IOERR;
		if (l != ID_SBOD) return ERROR_NOTMODULE;

		/* Read chunk size (length of instrument) & check it */
		if (FRead (fh, &size, 4, 1) != 1) return ERROR_IOERR;
		if (size != instr->Length) return ERROR_NOTMODULE;

		if (!(sample = AllocVec (instr->Length, MEMF_SAMPLE)))
			return ERROR_NO_FREE_STORE;

		/* Read instrument body */
		if (FRead (fh, sample, 1, instr->Length) != instr->Length)
		{
			FreeVec (sample);
			return ERROR_IOERR;
		}

		/* Double sample data */
		if (instr_mode[j] == OKT_MODE8 || instr_mode[j] == OKT_MODEB)
		{
			for (i = 0; i < instr->Length; i++)
				*sample++ <<= 1;
		}

		xmSetInstrument (si, j,
			INSTRA_Sample,	sample,
			TAG_DONE);
	}

	/* Check for extra data following the module */
	if (FGetC (fh) != ENDSTREAMCH)
		xmDisplayMessageA (XMDMF_NOTE | XMDMF_USECATALOG,
			(APTR)MSG_EXTRA_DATA_AFTER_MOD, NULL);

	return 0;
}



static HOOKCALL LONG SaveOktalyzer (
	REG(d0, BPTR fh),
	REG(a0, struct SongInfo *si),
	REG(a1, struct XMHook *saver),
	REG(a2, ULONG *tags))
{
	struct Instrument *instr;
	ULONG	i, j, k;
	UWORD	voices, instr_mode, songlen;
	ULONG	l;		/* Write buffers */
	UWORD	w;
	UBYTE	oktanote[4];

	/* Check number of tracks and fix data */
	voices = si->MaxTracks;
	if (voices < 4) voices = 4;
	if (voices > 8) voices = 8;


	/* Oktalyzer does not support pattern break (D) command */

	xmProcessSong (si, NULL,
		XMSNG_Optimize, XMOF_CUT_PATTERNS,
		TAG_DONE);


	/* Write file header */
	if (FWrite (fh, "OKTASONGCMOD", 12, 1) != 1) return ERROR_IOERR;


	/* Write maximum number of tracks */
	l = 8;
	if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	/* Write active tracks (only 5 to 8: tracks 1..4 are always on)
	 * TODO: Ask user what tracks should be made active, or loadnig and
	 * saving back the same module will result in the lost of the original
	 * track order.
	 */
	for (i = 5 ; i <= 8 ; i++)
	{
		if (voices >= i) w = 1;
			else w = 0;
		if (FWrite (fh, &w, 2, 1) != 1) return ERROR_IOERR;
	}

	/* Choose mode for instruments.
	 * When the module is 4 channels, we can always use mode 4.
	 * On modules with 5-7 channels, it is hard to guess which instruments
	 * could be made mode 4, 8 or B, so we always choose mode B.
	 * for 8 channels modules, we just use mode 8 for all instruments.
	 */
	switch (voices)
	{
		case 4:
			instr_mode = OKT_MODE4;
			break;

		case 8:
			instr_mode = OKT_MODE8;
			break;

		default:
			instr_mode = OKT_MODEB;
	}



	/* Write sample names, length, effects, volume */
	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_WRITING_INSTINFO, NULL);

	if (FWrite (fh, "SAMP", 4, 1) != 1) return ERROR_IOERR;

	/* Write chunk length */
	l = 0x480;
	if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	for ( j = 1 ; j <= 36 ; j++ )
	{
		struct
		{
			UBYTE	name[20];
			ULONG	size;
			UWORD	repeat,
					replen,
					volume,
					mode;
		} oktainstr = { 0 };


		if (instr = si->Instr[j])
		{
			strncpy (oktainstr.name, instr->Name, 19);
			oktainstr.size		= instr->Length;
			oktainstr.repeat	= instr->Repeat >> 1;
			oktainstr.replen	= instr->Replen >> 1;
			oktainstr.volume	= instr->Volume;
			oktainstr.mode		= instr->Length ? instr_mode : 0;
		}

		/* Write instrument data */
		if (FWrite (fh, &oktainstr, sizeof (oktainstr), 1) != 1)
			return ERROR_IOERR;
	}


	/* Write global song speed */

	if (FWrite (fh, "SPEE", 4, 1) != 1) return ERROR_IOERR;
	l = 2;
	if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	w = si->GlobalSpeed;
	if (FWrite (fh, &w, 2, 1) != 1) return ERROR_IOERR;


	if (FWrite (fh, "SLEN", 4, 1) != 1) return ERROR_IOERR;
	l = 2;
	if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	w = si->NumPatterns;
	if (FWrite (fh, &w, 2, 1) != 1) return ERROR_IOERR;


	/* Write patterns number */

	if (FWrite (fh, "PLEN", 4, 1) != 1) return ERROR_IOERR;
	l = 2;
	if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	songlen = min (si->Length, 128);
	if (FWrite (fh, &songlen, 2, 1) != 1) return ERROR_IOERR;


	/* Write patterns sequence */

	if (FWrite (fh, "PATT", 4, 1) != 1) return ERROR_IOERR;
	l = 128;
	if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

	{
		UBYTE postable[128];

		memset (postable, 0, 128);

		for (i = 0; i < songlen; i++)
			postable[i] = si->Sequence[i];

		if (FWrite (fh, postable, 1, 128) != 128) return ERROR_IOERR;
	}


	/* Write patterns */
	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_WRITING_PATTS, NULL);

	for ( j = 0 ; j < si->NumPatterns ; j++)
	{
		if (xmDisplayProgress (j + 1, si->NumPatterns))
			return ERROR_BREAK;

		if (FWrite (fh, "PBOD", 4, 1) != 1) return ERROR_IOERR;
		l = si->Patt[j]->Lines * si->MaxTracks * 4 + 2;
		if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

		/* Write pattern length (WORD) */
		w = si->Patt[j]->Lines;
		if (FWrite (fh, &w, 2, 1) != 1) return ERROR_IOERR;

		for (k = 0 ; k < si->Patt[j]->Lines ; k++)
		{
			for ( i = 0 ; i < voices ; i++)
			{
				struct Note *n = &(si->Patt[j]->Notes[i][k]);

				if (n->Note)
				{
					if (n->Note < 13)
					{
						xmDisplayMessage (XMDMF_NOTE | XMDMF_USECATALOG,
							(APTR)MSG_NOTE_TOO_LOW, j, i, k);
						oktanote[0] = n->Note;
					}
					else
					{
						if (n->Note > 48)
						{
							xmDisplayMessage (XMDMF_NOTE | XMDMF_USECATALOG,
								(APTR)MSG_NOTE_TOO_HIGH, j, i, k);
							oktanote[0] = n->Note - 24;
						}
						else
							oktanote[0] = n->Note - 12;
					}
				}
				else oktanote[0] = 0;

				oktanote[1] = (n->Inst ? (n->Inst-1) : 0);
				oktanote[2] = Effects[n->EffNum];

				/* Effect Exceptions */
				switch (n->EffNum)
				{
					case EFF_MISC:
						if ((n->EffVal >> 4) == 1)	/* Filter On/Off */
							oktanote[3] = (n->EffVal ? 0 : 1);
						break;

					case EFF_SETSPEED:
						if (n->EffVal > 0x0F)
							oktanote[3] = 0x0F;
						else
							oktanote[3] = n->EffVal;
						break;

					default:
						oktanote[3] = n->EffVal;
						break;
				}

				if (FWrite (fh, oktanote, 4, 1) != 1) return ERROR_IOERR;

			}	/* end for(i) */
		}	/* End for(k) */
	}	/* End for(j) */


	/* Write Instruments Data */

	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_WRITING_INSTDATA, NULL);

	for (j = 1 ; j <= 36 ; j++)
	{
		BYTE *samp;
		BOOL free_samp = FALSE;

		/* Skip empty instruments slots */

		if (!(instr = si->Instr[j]))
			continue;
		if (!(l = instr->Length))
			continue;

		if (xmDisplayProgress(j, 36))
			return ERROR_BREAK;


		if (instr_mode == OKT_MODE8 || instr_mode == OKT_MODEB)
		{
			if (samp = AllocVec (l, MEMF_ANY))
			{
				/* Halve volume */
				for (i = 0; i < l; i++)
					samp[i] = instr->Sample[i] >> 1;

				free_samp = TRUE;	/* Free this when done */
			}
			else
			{
				xmDisplayMessage (XMDMF_WARNING | XMDMF_USECATALOG,
					(APTR)MSG_NO_MEM_TO_HALVE, j);
				samp = instr->Sample;
			}
		}
		else samp = instr->Sample;


		if (!samp) l = 0;

		if (FWrite (fh, "SBOD", 4, 1) != 1) return ERROR_IOERR;
		if (FWrite (fh, &l, 4, 1) != 1) return ERROR_IOERR;

		if (l)
		{
			Flush (fh);
			if (Write (fh, samp, l) != l)
				return ERROR_IOERR;
		}

		if (free_samp) FreeVec (samp);
	}

	return 0;
}



HOOKCALL struct Library * _UserLibInit (REG(a6, struct Library *mybase))
{
	return mybase;
}



HOOKCALL struct Library * _UserLibCleanup (REG(a6, struct Library *mybase))
{
	return mybase;
}



HOOKCALL APTR _GetEngine (REG(a6, struct Library *mybase))
{
	return NULL;
}



HOOKCALL APTR _SetupXMHook (
	REG(a0, struct XModuleBase *_XModuleBase),
	REG(a6, struct Library *mybase))
{
	if (!(strcmp (_XModuleBase->xm_Library.lib_Node.ln_Name, "xmodule.library")))
	{
		SysBase			= *((struct ExecBase **)4);
		XModuleBase		= _XModuleBase;
		IntuitionBase	= XModuleBase->xm_IntuitionBase;
		DOSBase			= XModuleBase->xm_DOSBase;

		xmAddHook (
			XMHOOK_Type,			NT_XMLOADER,
			XMHOOK_Name,			(LONG)"Oktalyzer",
			XMHOOK_Priority,		0,
			XMHOOK_Descr,			(LONG)"Oktalyzer 1.57",
			XMHOOK_Author,			(LONG)"Bernardo Innocenti",
			XMHOOK_ID,				ID_OKTA,
			XMHOOK_LibraryBase,		mybase,
			XMHOOK_LoadModFunc,		LoadOktalyzer,
			XMHOOK_IdentifyModFunc,	IdentifyOktalyzer,
			XMHOOK_MaxTracks,		8,
			XMHOOK_MaxPatterns,		64,
			XMHOOK_MaxInstruments,	36,
			XMHOOK_MaxLength,		128,
			XMHOOK_MaxSampleLen,	131070,
			XMHOOK_MaxPattLen,		128,
			TAG_DONE);

		xmAddHook (
			XMHOOK_Type,			NT_XMSAVER,
			XMHOOK_Name,			(LONG)"Oktalyzer",
			XMHOOK_Priority,		0,
			XMHOOK_Descr,			(LONG)"Oktalyzer 1.57",
			XMHOOK_Author,			(LONG)"Bernardo Innocenti",
			XMHOOK_ID,				ID_OKTA,
			XMHOOK_LibraryBase,		mybase,
			XMHOOK_SaveModFunc,		SaveOktalyzer,
			XMHOOK_MaxTracks,		8,
			XMHOOK_MaxPatterns,		64,
			XMHOOK_MaxInstruments,	36,
			XMHOOK_MaxLength,		128,
			XMHOOK_MaxSampleLen,	131070,
			XMHOOK_MaxPattLen,		128,
			TAG_DONE);
	}

	return NULL;
}
