/*
**	ScreamTrackerHook.c
**
**	Copyright (C) 1995,96,97 Bernardo Innocenti
**
**	Load a ScreamTracker 3.01 module with any number of tracks.
**	Only sample instruments are supported.
*/


#include <exec/memory.h>
#include <libraries/xmodule.h>
#include <libraries/songclass.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/xmodule.h>

#include "XModulePriv.h"


/* Local function prototypes */

APTR HOOKCALL _GetEngine (
	REG(a6, struct Library *MyBase));
APTR HOOKCALL _SetupXMHook (
	REG(a0, struct XModuleBase *_XModuleBase),
	REG(a6, struct Library *mybase));
static HOOKCALL struct XMHook *IdentifyScreamTracker (
	REG(d0, BPTR fh),
	REG(a0, struct XMHook *loader),
	REG(a1, ULONG *tags));
static HOOKCALL LONG LoadScreamTracker (
	REG(d0, BPTR fh),
	REG(a0, struct SongInfo *si),
	REG(a1, struct XMHook *loader),
	REG(a2, ULONG *tags));
INLINE UBYTE DecodeEff (UBYTE eff, UBYTE effval);


#ifndef MAKE_ID
#define MAKE_ID(a,b,c,d) \
	((ULONG) (a)<<24 | (ULONG) (b)<<16 | (ULONG) (c)<<8 | (ULONG) (d))
#endif

/* Convert an Intel style WORD to Motorola format */
#define I2M(x) ( (UWORD) ( (((UWORD)(x)) >> 8) | (((UWORD)(x)) << 8) ) )

/* Convert an Intel style LONG to Motorola format */
#define I2ML(x) ( I2M((x)>>16) | (I2M((x))<<16) )

/* Convert a ParaPointer to a normal file offset */
#define PARA(x) (((ULONG)I2M(x)) << 4)

/* Convert a 3-byte ParaPointer to a normal file offset */
#define PARAL(x) (( (ULONG)x[0]<<16 | (ULONG)x[2]<<8 | (ULONG)x[1] ) << 4)


struct S3MHeader
{
	UBYTE SongName[28];
	UBYTE Constant;
	UBYTE Type;
	UWORD Pad0;
	UWORD OrdNum;		/* Number of positions */
	UWORD InsNum;
	UWORD PatNum;
	UWORD Flags;
	UWORD CreatedWith;
	UWORD FileFormat;	/* See below */
	ULONG ID;			/* Should be ID_SCRM */
	UBYTE GlobalVolume;
	UBYTE InitialSpeed;
	UBYTE InitialTempo;
	UBYTE MasterVolume;
	UBYTE Pad1[10];
	UWORD Special;
	UBYTE Channels[32];
};



/* Values for S3MHeader->FileFormat */

#define S3MFF_SIGNEDSAMPLES		1
#define S3MFF_UNSIGNEDSAMPLES	2



struct S3MSamp
{
	UBYTE Type;			/* See ITYPE_#? definitions below.	*/
	UBYTE DosName[12];
	UBYTE MemSeg[3];	/* :-))) Parapointer to sample data	*/
	ULONG Length;
	ULONG LoopBeg;
	ULONG LoopEnd;
	UBYTE Volume;
	UBYTE Pad0;
	UBYTE Pack;			/* See SPACK_#? definitions below. */
	UBYTE Flags;		/* See SFLAG_#? definitions below. */
	ULONG C2Spd;
	ULONG Pad1;
	UWORD Reserved0[2];
	ULONG Reserved1[1];
	UBYTE SampleName[28];
	ULONG SampID;		/* Should be ID_SCRS */
};


/* S3M Instrument types */

#define ITYPE_NONE			0
#define ITYPE_SAMPLE		1
#define ITYPE_ADLIB_MELODY	2
#define ITYPE_ADLIB_SNARE	3
#define ITYPE_ADLIB_TOM		4
#define ITYPE_ADLIB_CYMBAL	5
#define ITYPE_ADLIB_HIHAT	6


/* S3M Sample packing */

#define SPACK_NONE			0
#define SPACK_ADPCM			1	/* Unsupported by ST3.01 */


/* S3M Sample Flags */
#define SFLAG_LOOP			1
#define SFLAG_STEREO		2	/* Unsupported by ST3.01 */
#define SFLAG_16BIT			4	/* Unsupported by ST3.01 */


/* S3M IDs */

#define ID_S3M	MAKE_ID('S','3','M','\0')	/* ScreamTracker ID		*/
#define ID_SCRM	MAKE_ID('S','C','R','M')	/* ScreamTracker Module	*/
#define ID_SCRS	MAKE_ID('S','C','R','S')	/* ScreamTracker Sample	*/
#define ID_SCRI	MAKE_ID('S','C','R','I')	/* ScreamTracker Instr	*/



/* Effects conversion table */
static const UBYTE Effects[MAXTABLEEFFECTS] =
{
/*   S3M		XModule			Val */

	0x00,	/*	Null effect		$00	*/

	0x05,	/*	Portamento Up	$01	*/
	0x04,	/*	Portamento Down	$02	*/
	0x06,	/*	Tone Portamento	$03	*/
	0x07,	/*	Vibrato			$04	*/
	0x0B,	/*	ToneP + VolSl	$05	*/
	0x0A,	/*	Vibra + VolSl	$06	*/
	0x08,	/*	Tremolo			$07	*/
	0x00,	/*	Set Hold/Decay	$08	*/
	0x0E,	/*	Sample Offset	$09	*/
	0x03,	/*	Volume Slide	$0A	*/
	0x01,	/*	Position Jump  	$0B	*/
	0x12,	/*	Set Volume		$0C	*/
	0x02,	/*	Pattern break	$0D	*/
	0x10,	/*	Misc			$0E	*/
	0x00,	/*	Set Speed		$0F	*/
	0x11,	/*	Set Tempo		$10	*/
	0x09,	/*	Arpeggio		$11	*/

	0x03,	/*	Oktalyzer H        	*/
	0x03	/*	Oktalyzer L        	*/
};



/* Library data */

const UBYTE LibName[] = "screamtracker.xmhook";
const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
const UBYTE LibId[] = "screamtracker.xmhook 1.0 (17.3.96)  1995-96 by Bernardo Innocenti";


/* Get around a SAS/C bug which causes some annoying warnings
 * with the library bases defined below.  The problem happens
 * when the GST has been compiled with the CODE=FAR switch and
 * the source is being compiled with CODE=NEAR.
 */
#ifdef __SASC
#pragma msg 72 ignore push
#endif /* __SASC */

struct ExecBase			*SysBase		= NULL;
struct XModuleBase		*XModuleBase	= NULL;
struct DosLibrary		*DOSBase		= NULL;
struct IntuitionBase	*IntuitionBase	= NULL;

#ifdef __SASC
#pragma msg 72 ignore pop
#endif /* __SASC */


static HOOKCALL struct XMHook *IdentifyScreamTracker (
	REG(d0, BPTR fh),
	REG(a0, struct XMHook *loader),
	REG(a1, ULONG *tags))

/* Determine if the given file is a ScreamTracker 3.0 module.
 * Note: the file position will be changed on exit.
 */
{
	ULONG id;

	Seek (fh, 0x2C, OFFSET_BEGINNING);
	if (FRead (fh, &id, 4, 1) != 1)
		return NULL;

	return ((id == ID_SCRM) ? loader : NULL);
}



static HOOKCALL LONG LoadScreamTracker (
	REG(d0, BPTR fh),
	REG(a0, struct SongInfo *si),
	REG(a1, struct XMHook *loader),
	REG(a2, ULONG *tags))
{
	UWORD				*ParaInst,
						*ParaPatt;
	struct Note			*note;
	struct Pattern		*patt;
	struct S3MHeader	 s3mhd;

	ULONG		i, j;			/* Loop counters */
	LONG		err = 0;
	UWORD		numchannels = 0, w;
	LONG		l;				/* Dummy read buffer */


	/* Read module header */
	if (FRead (fh, &s3mhd, sizeof (s3mhd), 1) != 1)
		return ERROR_IOERR;

	/* Fix Intel WORD format */
	s3mhd.OrdNum		= I2M (s3mhd.OrdNum);
	s3mhd.InsNum		= I2M (s3mhd.InsNum);
	s3mhd.PatNum		= I2M (s3mhd.PatNum);
	s3mhd.Flags			= I2M (s3mhd.Flags);
	s3mhd.CreatedWith	= I2M (s3mhd.CreatedWith);
	s3mhd.FileFormat	= I2M (s3mhd.FileFormat);
	s3mhd.Special		= I2M (s3mhd.Special);


	for (i = 0; i < 32; i++)
		if (s3mhd.Channels[i] != 255) numchannels++;

	s3mhd.SongName[27] = '\0';	/* Ensure Null-termination */


	/* Read the pattern sequence */

	if (!xmSetSongLen (si, s3mhd.OrdNum))
		return ERROR_NO_FREE_STORE;

	for (i = 0; i < s3mhd.OrdNum; i++)
	{
		if ((l = FGetC (fh)) != -1L)
			si->Sequence[i] = l;
		else
			return ERROR_IOERR;
	}

	if (s3mhd.OrdNum & 1) FGetC (fh);	/* Keep WORD alignament */



	/*******************/
	/* Get Instruments */
	/*******************/

	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_READING_INSTS, NULL);

	/* Get parapointers to instruments */

	if (!(ParaInst = AllocVec (s3mhd.InsNum * sizeof (UWORD), MEMF_PUBLIC)))
		return ERROR_NO_FREE_STORE;

	if (!(ParaPatt = AllocVec (s3mhd.PatNum * sizeof (UWORD), MEMF_PUBLIC)))
	{
		FreeVec (ParaInst);
		return ERROR_NO_FREE_STORE;
	}

	if (FRead (fh, ParaInst, sizeof (UWORD), s3mhd.InsNum) == s3mhd.InsNum)
	{
		if (FRead (fh, ParaPatt, sizeof (UWORD), s3mhd.PatNum) == s3mhd.PatNum)
		{
			struct S3MSamp samp;

			/* Note: Need to Flush() here?  Hmm... I think not. */

			for (i = 0; i < s3mhd.InsNum; i++)
			{
				if (xmDisplayProgress (i + 1, s3mhd.InsNum))
				{
					err = ERROR_BREAK;
					break;
				}

				if (Seek (fh, PARA(ParaInst[i]), OFFSET_BEGINNING) == -1)
				{
					err = ERROR_IOERR;
					break;
				}
				else
				{
					struct Instrument *inst;
					BYTE *sample;

					if (Read (fh, &samp, sizeof (samp)) != sizeof (samp))
					{
						err = ERROR_IOERR;
						break;
					}

					samp.SampleName[27] = '\0';	/* Ensure NULL termination */

					if (samp.Type == 0)
						;	/* Do nothing */
					else if (samp.Type == ITYPE_SAMPLE)
					{
						if (samp.Pack != SPACK_NONE)
							xmDisplayMessage (XMDMF_ERROR | XMDMF_USECATALOG,
								(APTR)MSG_UNKNOWN_SAMPLE_COMPRESSION, i + 1);

						if (samp.Flags & SFLAG_STEREO)
							xmDisplayMessage (XMDMF_ERROR | XMDMF_USECATALOG,
							(APTR)MSG_INST_IS_STEREO, i + 1);

						if (samp.Flags & SFLAG_16BIT)
						{
							xmDisplayMessage (XMDMF_ERROR | XMDMF_USECATALOG,
								(APTR)MSG_INST_IS_16BIT, i + 1);
							continue;
						}

						if (!(inst = xmAddInstrument (si, i + 1,
							INSTRA_Name,	samp.SampleName,
							INSTRA_Length,	I2ML(samp.Length),
							INSTRA_Volume,	(samp.Volume > 64) ? 64 : samp.Volume,
							(samp.Flags & SFLAG_LOOP) ? INSTRA_LoopStart : TAG_IGNORE, I2ML(samp.LoopBeg),
							(samp.Flags & SFLAG_LOOP) ? INSTRA_LoopEnd : TAG_IGNORE,	I2ML(samp.LoopEnd),
							TAG_DONE)))
							return ERROR_NO_FREE_STORE;


						/* Seek where the sample is stored */
						if (Seek (fh, PARAL(samp.MemSeg), OFFSET_BEGINNING) == -1)
						{
							err = ERROR_IOERR;
							break;
						}

						/* Allocate memory for sample */

						if (!(sample = AllocVec (inst->Length, MEMF_SAMPLE)))
						{
							err = ERROR_NO_FREE_STORE;
							break;
						}

						/* Load sample data */

						if (Read (fh, sample, inst->Length) != inst->Length)
						{
							FreeVec (sample);
							err = ERROR_IOERR;
							break;
						}

						if (s3mhd.FileFormat >= S3MFF_UNSIGNEDSAMPLES)
						{
							/* unsigned -> signed  conversion */
							for (j = 0; j < inst->Length; j++)
								sample[j] ^= 0x80;
						}

						xmSetInstrument (si, i,
							INSTRA_Sample,	sample,
							TAG_DONE);
					}
					else if (samp.Type <= ITYPE_ADLIB_HIHAT)
					{
						static UBYTE *ADLibNames[] =
						{
							"Melody",
							"Snare Drum",
							"Tom",
							"Cymbal",
							"Hihat"
						};

						xmDisplayMessage (XMDMF_ERROR | XMDMF_USECATALOG,
							(APTR)MSG_ADLIB_INSTR, i + 1, ADLibNames[samp.Type - 2]);
					}
					else
						xmDisplayMessage (XMDMF_ERROR | XMDMF_USECATALOG,
							(APTR)MSG_ERR_NOT_A_SAMPLE, i + 1);
				}
			}
		}
		else err = ERROR_IOERR;
	}
	else err = ERROR_IOERR;

	FreeVec (ParaInst);

	SetAttrs (si,
		SNGA_Title,			s3mhd.SongName,
		SNGA_Author,		-1,
		SNGA_GlobalSpeed,	s3mhd.InitialSpeed,
		SNGA_GlobalTempo,	s3mhd.InitialTempo,
		TAG_DONE);

	if (err)
	{
		FreeVec (ParaPatt);
		return err;
	}


	/****************/
	/* Get Patterns */
	/****************/

	xmDisplayMessageA (XMDMF_INFORMATION | XMDMF_ACTION | XMDMF_USECATALOG,
		(APTR)MSG_READING_PATTS, NULL);

	for (i = 0; i < s3mhd.PatNum; i++)
	{
		if (xmDisplayProgress (i + 1, s3mhd.PatNum))
		{
			err = ERROR_BREAK;
			break;
		}

		if (Seek (fh, PARA(ParaPatt[i]), OFFSET_BEGINNING) == -1)
		{
			err = ERROR_IOERR;
			break;
		}
		else
		{
			LONG ch = 0;	/* Channel number */

			/* Read size */

			if (Read (fh, &w, sizeof (UWORD)) != sizeof (UWORD))
			{
				err = ERROR_IOERR;
				break;
			}

			if (patt = xmAddPattern (si,
				PATTA_Tracks,	numchannels,
				PATTA_Lines,	64,
				TAG_DONE))
			{
				/* Loop on rows */
				for (j = 0; (j < 64) && (ch != -1L); j++)
				{
					while ((ch = FGetC (fh)) != -1L)
					{
						if (ch == 0)	/* End of row */
							break;

						if ((ch & 31) >= numchannels)
						{
							xmDisplayMessage (XMDMF_WARNING | XMDMF_USECATALOG,
								(APTR)MSG_TRACK_OUT_OF_RANGE, ch & 31);
							note = &patt->Notes[0][j];
						}
						else
							note = &patt->Notes[ch & 31][j];

						if (ch & 32)	/* Note and instrument follows */
						{
							l = FGetC (fh);
							if (l != 255)
								note->Note = ((((l >> 4) - 2) * 12) | (l & 0x0F)) + 1;
							note->Inst = FGetC (fh);
						}
						if (ch & 64)	/* Volume */
						{
							l = FGetC (fh);
							note->EffNum = EFF_SETVOLUME;
							note->EffVal = l;
						}

						if (ch & 128)	/* Command and Info */
						{
							note->EffNum = FGetC (fh);
							note->EffVal = FGetC (fh);
							note->EffNum = DecodeEff (note->EffNum, note->EffVal);
						}
					}
				}

				if (ch == -1)
				{
					err = ERROR_IOERR;
					break;
				}
			}
			else
			{
				err = ERROR_NO_FREE_STORE;
				break;
			}

			/* Flush (fh); */
		}
	}

	FreeVec (ParaPatt);

	return err;
}



INLINE UBYTE DecodeEff (UBYTE eff, UBYTE effval)
{
	UBYTE i;

	if ((eff == 0) && effval) /* Speed/Tempo */
	{
		if (effval < 0x20)
			return EFF_SETSPEED;
		else
			return EFF_SETTEMPO;
	}

	for (i = 0 ; i < MAXTABLEEFFECTS; i++)
		if (eff == Effects[i])
			return i;

	return 0;
}



HOOKCALL struct Library * _UserLibInit (REG(a6, struct Library *mybase))
{
	return mybase;
}



HOOKCALL struct Library * _UserLibCleanup (REG(a6, struct Library *mybase))
{
	return mybase;
}



HOOKCALL APTR _GetEngine (REG(a6, struct Library *mybase))
{
	return NULL;
}



HOOKCALL APTR _SetupXMHook (
	REG(a0, struct XModuleBase *_XModuleBase),
	REG(a6, struct Library *mybase))
{
	if (!(strcmp (_XModuleBase->xm_Library.lib_Node.ln_Name, "xmodule.library")))
	{
		SysBase = *((struct ExecBase **)4);
		XModuleBase		= _XModuleBase;
		IntuitionBase	= XModuleBase->xm_IntuitionBase;
		DOSBase			= XModuleBase->xm_DOSBase;

		xmAddHook (
			XMHOOK_Type,			NT_XMLOADER,
			XMHOOK_Name,			(LONG)"ScreamTracker",
			XMHOOK_Priority,		0,
			XMHOOK_Descr,			(LONG)"ScreamTracker 3.01",
			XMHOOK_Author,			(LONG)"Bernardo Innocenti",
			XMHOOK_ID,				ID_S3M,
			XMHOOK_Flags,			0,
			XMHOOK_LibraryBase,		mybase,
			XMHOOK_LoadModFunc,		LoadScreamTracker,
			XMHOOK_IdentifyModFunc,	IdentifyScreamTracker,
			XMHOOK_MaxTracks,		16,
			XMHOOK_MaxPatterns,		100,
			XMHOOK_MaxInstruments,	99,
			XMHOOK_MaxLength,		255,
			XMHOOK_MaxSampleLen,	64000,
			XMHOOK_MaxPattLen,		64,
			TAG_DONE);
	}

	return NULL;
}
