#ifndef LIBRARIES_PATTEDITCLASS_H
#define LIBRARIES_PATTEDITCLASS_H
/*
**	PattEditClass.h
**
**	Copyright (C) 1995 by Bernardo Innocenti
**
**	Pattern editor class built on top of the "gadgetclass".
**
**	Note: Use 4 chars wide TABs to read this file.
*/

#ifndef DATATYPES_DATATYPESCLASS_H
#include <datatypes/datatypesclass.h>
#endif


#define PATTEDITNAME	"gadgets/pattedit.gadget"
#define PATTEDITCLASS	"patteditclass"
#define PATTEDITVERS	2


/********************/
/* Class Attributes */
/********************/

#define PEA_Dummy				(TAG_USER + 'PE' << 8)

#define PEA_CursTrack			(PEA_Dummy+1)
#define PEA_CursColumn			(PEA_Dummy+2)
#define PEA_CursLine			(PEA_Dummy+3)
	/* (IGSNU) Cursor position.
	 */

#define PEA_LeftTrack			DTA_TopHoriz
#define PEA_TopLine				DTA_TopVert
	/* (IGSNU) Top line and leftmost track of the editor view.
	 */

#define PEA_Left				(PEA_Dummy+6)
#define PEA_Right				(PEA_Dummy+7)
#define PEA_Up					(PEA_Dummy+8)
#define PEA_Down				(PEA_Dummy+9)
#define PEA_CursLeft			(PEA_Dummy+10)
#define PEA_CursRight			(PEA_Dummy+11)
#define PEA_CursUp				(PEA_Dummy+12)
#define PEA_CursDown			(PEA_Dummy+13)
	/* (S) PEA_Left and PEA_Right scroll the view left and right one
	 * line.  The cursor is positioned to the leftmost or rightmost
	 * position, respectively. PEA_Up and PEA_Down scroll the view up
	 * and down of one line.  The cursor is positioned to the upper or
	 * lower visible line, respectively.   PEA_Curs#? attributes move
	 * the cursor one position towards the direction specified.  The
	 * contents of ti_Data is meaningless for all these attributes.
	 */

#define PEA_UndoChange			(PEA_Dummy+14)
	/* (S) If ti_Data is > 0, one change made to the pattern is
	 * undone.  If ti_Data is < 0, a previously undone change is
	 * redone.  The cursor is always moved on the affected position.
	 */

#define PEA_Changes				(PEA_Dummy+15)
	/* (IGS) ti_Data contains the number of changes made to the pattern.
	 */

#define PEA_MarkRegion			(PEA_Dummy+16)
	/* (GS) ti_Data points to a struct Rectangle containing the limits
	 * (tracks and lines) of a pattern sub-region.  When set, this
	 * attribute automatically starts range mode.  Passing NULL in
	 * ti_Data causes the editor to clear the marked region.  Passing
	 * ~0 in ti_Data toggles mark mode on and off.
	 * Getting this attribute returns a copy of the currently
	 * marked region, or {0,0,0,0} if no region is currently
	 * selected.  The following relations must be true when setting
	 * this attribute and are guaranteed when getting it:
	 *
	 * 	MinX <= MaxX
	 *	MinY <= MaxY
	 *	MaxX < Patt->Tracks
	 *	MaxY < Patt->Lines
	 */

#define PEA_Flags				(PEA_Dummy+17)
	/* (IGSN) See PEF_#? flags definitions below.
	 */

#define PEA_DisplayTracks		DTA_VisibleHoriz
#define PEA_DisplayLines		DTA_VisibleVert
	/* (GN) Maximum number of tracks and lines that fit in the gadget
	 * bounds.
	 */

#define PEA_Pattern			(PEA_Dummy+20)
	/* (ISGU) ti_Data is a pointer to the Pattern
	 * structure to be displayed by the PatternEditor.
	 */

#define PEA_CurrentInst		(PEA_Dummy+21)
	/* (IS) ti_Data is the default instrument
	 * number to be used when entering notes in the
	 * pattern editor.
	 */

#define PEA_MaxUndoLevels		(PEA_Dummy+22)
	/* (IS) ti_Data is the maximum size of the undo buffer in number
	 * of slots.  0 disables undo feature.  Defaults to 16.
	 */

#define PEA_MaxUndoMem			(PEA_Dummy+23)
	/* (IS) ti_Data is the maximum memory used by the undo buffers.
	 * Setting it to 0 means unlimited memory.  Defaults to 8192 bytes.
	 */

#define PEA_TextFont			DTA_TextFont
	/* (I) ti_Data points to the TextFont to be
	 * used with the PatternEditor.  The font must
	 * be mono spaced.
	 */

#define PEA_AdvanceCurs			(PEA_Dummy+25)
	/* (IS) The lower 16 bits of ti_Data contain the number of lines
	 * the cursor moves when a note is typed. The upper 16 bits contain
	 * the number of tracks. Both numbers are signed, so backwards
	 * cursor advancement is possible.
	 */

#define PEA_CursWrap			(PEA_Dummy+26)
	/* (IS) Can be 0 for no wrapping, PEF_HWRAP, PEF_VWRAP or both.
	 */

#define PEA_BGPen				(PEA_Dummy+27)
#define PEA_TextPen				(PEA_Dummy+28)
#define PEA_LinesPen			(PEA_Dummy+29)
#define PEA_TinyLinesPen		(PEA_Dummy+30)
	/* (IS) Pens to be used to render the various editor elements.
	 * PEA_TextPen must be a power of two (1, 2, 4, 8...) because
	 * the text is rendered on one bitplane only. PEA_LinesPen
	 * and PEA_TinyLinesPen must not have the PEA_TextPen bit
	 * set because text scrolling happens on the text bitplane only
	 * and must not disturb the other elements.  The defaults are
	 * 1 for PEA_TextPen and 2 for both PEA_LinesPen and
	 * PEA_TinyLinesPen.
	 */

#define PEA_KeyboardMap		(PEA_Dummy+31)
	/* (IS) ti_Data points to a table which describes the notes associated
	 * to of the keys.  The table starts with 1 Word specifying the number
	 * of keys following.
	 *
	 * Each key definition has this format:
	 *
	 *	1 Byte with the note associated with the first key.
	 *	1 Byte with the instrument associated with the first key.
	 *		This value may be 0, in which case the current instrument
	 *		will be used.
	 *
	 * Passing NULL in ti_Data disables keyboard mapping.
	 *
	 * NOTE: The table is referenced, not copied, so it must stay in
	 * memory until the pattern editor is disposed or the table is
	 * disabled.
	 */

#define PEA_TrackChars		(PEA_Dummy+32)
	/* (IGSNU) Number of straight characters in a track.  Minimum
	 * value is MINTRACKCHARS, maximum is MAXTRACKCHARS.
	 * Defaults to 10.
	 */

#define PEA_ChangeNote		(PEA_Dummy+33)
	/* (GNU) ti_Data contains a packed track/line pair of coordinates
	 * (the line in the low order 16 bits and the track in the upper
	 * 16bits). The editor sends update messages containing this tag
	 * whenever a note has been changed, and redraws the specified note
	 * when it receives this tag.  This is a way to keep two or more
	 * editors synchronized while they display the same pattern.  It can
	 * also be used to count the changes made to the song.
	 * ti_Data will be -1L when more than one note has changed (for
	 * istance when the user has pasted a block), and other editors
	 * will have to redraw their display completely.
	 */

#define PEA_SongInfo		(PEA_Dummy+34)
	/* (ISGU) ti_Data is a pointer to the songclass object which
	 * contains the pattern to be displayed.  Without this tag, the
	 * pattern editor is unable to do some operations.
	 */

#define PEA_PattNum			(PEA_Dummy+35)
	/* (ISGNU) The pattern number to display. To use this tag,
	 * the editor needs to be passed the PEA_SongInfo tag too.
	 */


/* Definitions for PEA_Flags attribute */

#define PEF_MARKING			(1<<0)	/* Range mode						*/
#define PEF_HWRAP			(1<<1)	/* Horizontal cursor wrap			*/
#define PEF_VWRAP			(1<<2)	/* Vertical cursor wrap				*/
#define PEF_HEXMODE			(1<<3)	/* Use hexadecimal numbers			*/
#define PEF_BLANKZERO		(1<<4)	/* Blank zero digits				*/
#define PEF_INVERSETEXT		(1<<5)	/* Show backfilled text				*/
#define PEF_DOTINYLINES		(1<<6)	/* Show tiny separator lines		*/
#define PEF_DOCURSORRULER	(1<<7)	/* Show a ruler under the cursor	*/
#define PEF_MARKFULLTRACKS	(1<<8)	/* Mark full tracks only			*/

/* Private flags - Not settable by application */
#define PEF_ADJUSTSCROLLBOX	(1<<28)	/* User is adjusting the scroll box	*/
#define PEF_RESIZETRACKS	(1<<29)	/* User is resizing the tracks		*/
#define PEF_SCROLLING		(1<<30)	/* View is scrolling (read only)	*/
#define PEF_DRAGGING		(1<<31)	/* Cursor drag mode (read only)		*/



/* Width of a track expressed in chars */
#define MAXTRACKCHARS	14
#define MINTRACKCHARS	4


/* Cursor column names */
enum {
	COL_NOTE,	/* Cursor on note field					*/
	COL_INSTH,	/* Cursor on instrument high nibble		*/
	COL_INSTL,	/* Cursor on instrument low  nibble 	*/
	COL_EFF,	/* Cursor on effect field				*/
	COL_VALH,	/* Cursor on effect value high nibble	*/
	COL_VALL,	/* Cursor on effect value low  nibble	*/

	COL_COUNT
};

#endif /* !LIBRARIES_PATTEDITCLASS_H */
