#ifndef SAMPEDITCLASS_H
#define SAMPEDITCLASS_H
/*
**	SampEditClass.h
**
**	Copyright (C) 1995,96 by Bernardo Innocenti
**
**	Sample editor class built on top of the "gadgetclass".
**
**	Note: Use 4 chars wide TABs to read this file.
*/


#define SAMPEDITNAME	"gadgets/sampedit.gadget"
#define SAMPEDITCLASS	"sampeditclass"
#define SAMPEDITVERS	1


/********************/
/* Class Attributes */
/********************/

#define SEA_CursXPos			(TAG_USER+1)
#define SEA_CursYPos			(TAG_USER+2)
	/* (IGSNU) Cursor position. CursYPos is unused.
	 */

#define SEA_XStart			(TAG_USER+2)
#define SEA_YStart			(TAG_USER+3)
	/* (IGSNU) Left and top offsett of the display relative to the start of the sample.
	 */

#define SEA_XRatio			(TAG_USER+4)
#define SEA_YRatio			(TAG_USER+5)
	/* (ISGNU) Fixed point 16+16 bit number for the zoom factor. The upper word holds
	 * the integer part and the low word represents the decimals.  Setting the ratio
	 * to 65536 sets the zoom factor to 1:1.  Bigger numbers produce an higher zoom
	 * factor (e.g.: 131072 is 2:1).
	 */

#define SEA_ScrollLeft		(TAG_USER+6)
#define SEA_ScrollRight		(TAG_USER+7)
#define SEA_ScrollUp		(TAG_USER+8)
#define SEA_ScrollDown		(TAG_USER+9)
#define SEA_CursLeft		(TAG_USER+10)
#define SEA_CursRight		(TAG_USER+11)
#define SEA_CursUp			(TAG_USER+12)
#define SEA_CursDown		(TAG_USER+13)
	/* (S) SEA_ScrollLeft and SEA_ScrollRight scroll the display left
	 * and right one unit.  SEA_ScrollUp and SEA_ScrollDown move the
	 * display up and down of one unit. SEA_Curs#? attributes move
	 * the cursor one position towards the direction specified.  The
	 * contents of ti_Data is meaningless for all these attributes.
	 */

#define SEA_UndoChange		(TAG_USER+14)
	/* (S) If ti_Data is > 0, one change made to the pattern is
	 * undone.  If ti_Data is < 0, a previously undone change is
	 * redone.  The cursor is always moved on the affected position.
	 */

#define SEA_Changes			(TAG_USER+15)
	/* (IGS) ti_Data contains the number of changes made to the sample.
	 */

#define SEA_MarkRegion		(TAG_USER+16)
	/* (GS) ti_Data points to a struct Rectangle containing the
	 * limits (tracks and lines) of a pattern sub-region.
	 * When set, this attribute automatically starts range mode.
	 * Passing NULL in ti_Data causes the editor to clear the
	 * marked region. Passing ~0 in ti_Data toggles mark mode.
	 * Getting this attribute returns a copy of the currently
	 * marked region, or {0,0,0,0} if no region is currently
	 * selected.
	 */

#define SEA_Flags			(TAG_USER+17)
	/* (IGSN) See PEF_#? flags definitions below.
	 */

#define SEA_Instrument	(TAG_USER+18)
	/* (ISGU) ti_Data is a pointer to the Instrument
	 * structure to be displayed by the SampleEditor.
	 */

#define SEA_MaxUndoLevels		(TAG_USER+19)
	/* (IS) ti_Data is the maximum size of the undo buffer in number
	 * of slots.  0 disables undo feature.  Defaults to 8.
	 */

#define SEA_MaxUndoMem		(TAG_USER+20)
	/* (IS) ti_Data is the maximum memory used by the undo buffers.
	 * Setting it to 0 means unlimited memory.  Defaults to 32768 bytes.
	 */

#define SEA_BackgroundPen	(TAG_USER+21)
#define SEA_LinePen		(TAG_USER+22)
#define SEA_FillPen		(TAG_USER+23)
#define SEA_GridPen		(TAG_USER+24)
	/* (IS) Pens to be used to render the various editor elements.
	 * SEA_TextPen must be a power of two (1, 2, 4, 8...) because
	 * the text is rendered on one bitplane only. SEA_LinesPen
	 * and SEA_TinyLinesPen must not have the SEA_TextPen bit
	 * set because text scrolling happens on the text bitplane only
	 * and must not disturb the other elements.  The defaults are
	 * 1 for SEA_TextPen and 2 for both SEA_LinesPen and
	 * SEA_TinyLinesPen.
	 */



/* Definitions for SEA_Flags attribute */

#define SEF_MARKING			(1<<0)	/* Range mode						*/
#define SEF_LOGMODE			(1<<1)	/* Logaritmic mode on				*/
#define SEF_DOGRID			(1<<2)	/* Show superimposed grid			*/

/* Private flags - Not settable by application */
#define SEF_SCROLLING		(1<<30)	/* View is scrolling (read only)	*/
#define SEF_DRAGGING		(1<<31)	/* Cursor drag mode (read only)		*/


#endif /* SAMPEDITCLASS_H */
