#ifndef LIBRARIES_XMODULE_H
#define LIBRARIES_XMODULE_H
/*
**	XModule.h
**
**	Copyright (C) 1995,96,97 Bernardo Innocenti
**
**	Set TAB size to 8 chars to read this header file
**
**	Public structures and constants for the xmodule.library
*/

#ifndef	DOS_DOS_H
#include <dos/dos.h>
#endif /* DOS_DOS_H */

#ifndef	EXEC_LISTS_H
#include <exec/lists.h>
#endif /* EXEC_LISTS_H */

#ifndef	EXEC_LIBRARIES_H
#include <exec/libraries.h>
#endif /* EXEC_LIBRARIES_H */

#ifndef	EXEC_SEMAPHORES_H
#include <exec/semaphores.h>
#endif /* EXEC_SEMAPHORES_H */

#ifndef	INTUITION_CLASSES_H
#include <intuition/classes.h>
#endif /* INTUITION_CLASSES_H */

#ifndef COMPILERSPECIFIC_H
#include "CompilerSpecific.h"
#endif /* COMPILERSPECIFIC_H */


/***************/
/* XModuleBase */
/***************/

struct XModuleBase
{
	struct Library			 xm_Library;		/* Standard Library node structure	*/
	UWORD					 xm_Pad0;			/* Struct Library is misaligned!	*/
	struct SignalSemaphore	 xm_BaseLock;		/* Obtain this before browsing
												 * in any of the fields!
												 */
	UWORD					 xm_Pad1;			/* SignalSemaphore is misaligned!	*/

	/* The following data is longword aligned */

	struct MinList			 xm_Songs;			/* List of public SongInfo structs	*/
	struct MinList			 xm_Loaders;		/* List of available loaders		*/
	struct MinList			 xm_Savers;			/* List of available savers			*/

	void					*xm_Pool;			/* Public mem pool used by XModule	*/
	Class					*xm_SongClass;		/* Song boopsi class				*/
	struct SongInfo			*xm_CurrentSong;	/* Pointer to currently active song	*/
	struct XMHook			*xm_DefaultSaver;	/* Currently selected saver hook	*/
	struct Screen			*xm_Screen;			/* XModule screen (might be NULL)	*/
	ULONG					 xm_Flags;			/* See definitions below			*/
	STRPTR					 xm_ScreenName;		/* Name of XModule's public screen	*/
	STRPTR					 xm_RexxName;		/* Name of XModule's ARexx port		*/

	ULONG					 xm_Reserved[4];

	struct DosLibrary		*xm_DOSBase;		/* Some useful library bases that	*/
	struct UtilityBase		*xm_UtilityBase;	/* can be grabbed and used by		*/
	struct Library			*xm_IFFParseBase;	/* external hooks.  These four are	*/
	struct IntuitionBase	*xm_IntuitionBase;	/* guaranteed to be always open.	*/
};



struct XMHook
{
	struct Node	xmh_Link;					/* Link to other hooks in a list.
											 * ln_Name contains the Hook name.
											 */
	UWORD		xmh_Version;				/* Will be set to XMHOOK_VERSION	*/
	STRPTR		xmh_Descr;					/* Short description of this hook	*/
	STRPTR		xmh_Author;					/* Author's name					*/
	ULONG		xmh_Flags;					/* See below for definitions		*/
	ULONG		xmh_ID;						/* IFF-style ID.  Set it to the FORM
											 * type for IFF loaders/savers.
											 */
	APTR		xmh_UserData;				/* User's private data				*/
	struct Library *xmh_LibraryBase;		/* External hook's library base		*/
	ASMCALL void	(*xmh_RemoveHook) (
		REG(a0,struct XMHook *hook));		/* Hook remotion entrypoint			*/
	ASMCALL LONG	(*xmh_LoadMod) (
		REG(d0, BPTR fh),
		REG(a0, struct SongInfo *song),
		REG(a1, struct XMHook *loader),
		REG(a2, struct TagItem *tags));		/* Module loader entrypoint			*/
	ASMCALL LONG	(*xmh_SaveMod) (
		REG(d0, BPTR fh),
		REG(a0, struct SongInfo *song),
		REG(a1, struct XMHook *saver),
		REG(a2, struct TagItem *tags));		/* Module saver entrypoint			*/
	ASMCALL struct XMHook * (*xmh_IdentifyMod) (
		REG(d0, BPTR fh),
		REG(a0, struct XMHook *hook),
		REG(a1, struct TagItem *tags));		/* Module identification entrypoint	*/

	ULONG		xmh_MaxTracks;				/* Max. number of tracks			*/
	ULONG		xmh_MaxPatterns;			/* Max. number of patterns			*/
	ULONG		xmh_MaxInstruments;			/* Max. number of instruments		*/
	ULONG		xmh_MaxLength;				/* Max. number of positions			*/
	ULONG		xmh_MaxSampleLen;			/* Max. length of each sample		*/
	ULONG		xmh_MaxPattLen;				/* Max. lines in a pattern			*/

	ULONG		xmh_Reserved[2];			/* Avoid recompilation :-)			*/
};



#define XMHOOK_VERSION		0

#define NT_XMSONG		'X'
#define NT_XMLOADER		'L'
#define NT_XMSAVER		'S'



/* Flag definitions for XMHook->xmh_Flags */

/* Basic type */
#define XMHF_BINARY					0x0000
#define XMHF_ASCII					0x0001
#define XMHF_IFF					0x0002
#define XMHF_MISC					0x0003

/* These specify that the saver is able to exclude
 * the corresponding items when saving in its own format.
 */
#define XMHF_EXCLUDE_SEQUENCE		(1<<5)
#define XMHF_EXCLUDE_INSTRUMENTS	(1<<6)
#define XMHF_EXCLUDE_PATTERNS		(1<<7)
#define XMHF_EXCLUDE_NAMES			(1<<8)

/* This module format does not support variable length
 * patterns.  Use in conjunction with XMHOOK_MaxPattLen tag.
 */
#define XMHF_FIXED_PATT_LEN			(1<<9)

/* This is an internal hook */
#define XMHF_INTERNAL	(1<<31)



/* Flags for xmDisplayMessage() */



#define XMDMF_ALERT			0x00000000
#define XMDMF_FAIL			0x00000001
#define XMDMF_ERROR			0x00000002
#define XMDMF_WARNING		0x00000003
#define XMDMF_NOTE			0x00000004
#define XMDMF_INFORMATION	0x00000005
#define XMDMF_COMMENT		0x00000006
#define XMDMF_DEBUG			0x00000007
#define XMDMF_LEVELMASK		0x00000007

#define XMDMF_DOSERRORMASK	0x0000FFFF
#define XMDMF_DOSFAULT		(1<<28)
#define XMDMF_ACTION		(1<<29)
#define XMDMF_USEREQUESTER	(1<<30)
#define XMDMF_USECATALOG	(1<<31)



/* Specific error codes */

#define ERROR_NOTMODULE	ERROR_OBJECT_WRONG_TYPE	/* File format not recognized		*/
#define ERROR_IOERR		RETURN_FAIL				/* A DOS call (e.g.: Read()) failed.
												 * check IoErr() for specific error.
												 */


/********/
/* Tags */
/********/

#define XMSNG_Dummy				(TAG_USER + 'XS'<<8)

#define XMSNG_AddToList			(XMSNG_Dummy + 1)
#define XMSNG_OldSong			(XMSNG_Dummy + 2)
#define XMSNG_Loader			(XMSNG_Dummy + 3)
#define XMSNG_FileHandle		(XMSNG_Dummy + 4)
#define XMSNG_Active			(XMSNG_Dummy + 5)

/* Tags for xmOptimizeSongA() */

#define XMSNG_Optimize			(XMSNG_Dummy + 16)
#define XMSNG_RemapInstruments	(XMSNG_Dummy + 17)
#define XMSNG_LimitPatterns		(XMSNG_Dummy + 18)
#define XMSNG_Join				(XMSNG_Dummy + 19)
#define XMSNG_Merge				(XMSNG_Dummy + 20)

/* Flags for XMSNG_Optimize */
#define XMOF_DEFAULT				(1<<0)
#define XMOF_REM_UNUSED_PATTERNS	(1<<1)
#define XMOF_REM_DUP_PATTERNS		(1<<2)
#define XMOF_REM_UNUSED_INSTRS		(1<<3)
#define XMOF_REM_DUP_INSTRS			(1<<4)
#define XMOF_CUT_INSTR_LOOPS		(1<<5)
#define XMOF_CUT_INSTR_TAILS		(1<<6)
#define XMOF_CUT_PATTERNS			(1<<7)
#define XMOF_REMAP_INSTRS			(1<<8)


/* XModule hook tags */

#define XMHOOK_Dummy			(TAG_USER + 'XH'<<8)

#define XMHOOK_Type				(XMHOOK_Dummy + 1)
#define XMHOOK_Name				(XMHOOK_Dummy + 2)
#define XMHOOK_Priority			(XMHOOK_Dummy + 3)
#define XMHOOK_Descr			(XMHOOK_Dummy + 4)
#define XMHOOK_Author			(XMHOOK_Dummy + 5)
#define XMHOOK_ID				(XMHOOK_Dummy + 6)
#define XMHOOK_Flags			(XMHOOK_Dummy + 7)
#define XMHOOK_UserData			(XMHOOK_Dummy + 8)
#define XMHOOK_LibraryBase		(XMHOOK_Dummy + 9)
#define XMHOOK_RemoveFunc		(XMHOOK_Dummy + 10)
#define XMHOOK_LoadModFunc		(XMHOOK_Dummy + 11)
#define XMHOOK_SaveModFunc		(XMHOOK_Dummy + 12)
#define XMHOOK_IdentifyModFunc	(XMHOOK_Dummy + 13)
#define XMHOOK_MaxTracks		(XMHOOK_Dummy + 14)
#define XMHOOK_MaxPatterns		(XMHOOK_Dummy + 15)
#define XMHOOK_MaxInstruments	(XMHOOK_Dummy + 16)
#define XMHOOK_MaxLength		(XMHOOK_Dummy + 17)
#define XMHOOK_MaxSampleLen		(XMHOOK_Dummy + 18)
#define XMHOOK_MaxPattLen		(XMHOOK_Dummy + 19)


#endif /* !LIBRARIES_XMODULE_H */
